/*	Filename:	columns.cc

	Created	December 14, 2006	Brandon Fosdick <bfoz@bfoz.net>

	Copyright Terran Development Co 2005 (BSD License)

	$Id: columns.cc,v 1.7 2007/04/16 05:58:11 bfoz Exp $
*/

#include "http_protocol.h"
#include "http_core.h"

#include "apr_strings.h"

#include "mod_log_dbd.h"
#include "columns.h"

namespace log_dbd
{
	const char allFormatSpecifiers[] = "AabcfHhilmPpRrSsTtUuVv";
	const char combinedFormatSpecifiers[] = "tBsAHhlmRuUd";	// NCSA "combined" equivalent
	const char combinedIOFormatSpecifiers[] = "tBsAHhlmRuUdIO";	// NCSA "combined IO" equivalent

	static const char* getAgent(request_rec* r)
	{
		const char* tempag = apr_table_get(r->headers_in, "User-Agent");
		return tempag ? tempag : "";
	}

	static const char* getRemoteIP(request_rec* r)
	{
		return r->connection->remote_ip;
	}

	static const char* getRequestArgs(request_rec* r)
	{
		return (r->args) ? r->args : "";
	}

	static const char* getBytesSent(request_rec* r)
	{
		return apr_off_t_toa(r->pool, r->bytes_sent);
	}

	static const char* getElapsedSeconds(request_rec* r)
	{
		apr_time_t duration = apr_time_now() - r->request_time;
		return apr_psprintf(r->pool, "%" APR_TIME_T_FMT, apr_time_sec(duration));
	}

	static const char* getElapsedMicroSeconds(request_rec* r)
	{
		return apr_psprintf(r->pool, "%" APR_TIME_T_FMT, (apr_time_now() - r->request_time));
	}

	static const char* getRequestFile(request_rec* r)
	{
		return r->filename;
	}

	static const char* getRequestProtocol(request_rec* r)
	{
		return r->protocol;
	}

	static const char* getRemoteHost(request_rec* r)
	{
		return ap_get_remote_host(r->connection, r->per_dir_config, REMOTE_NAME, NULL);
	}

	static const char* getUniqueID(request_rec* r)
	{
		const char *tempid = apr_table_get(r->subprocess_env, "UNIQUE_ID");
		return tempid ? tempid : "";
	}

	static const char* getRemoteLogname(request_rec* r)
	{
		const char* t = ap_get_remote_logname(r);
		return (NULL==t) ? "" : t;
	}

	static const char* getRequestMethod(request_rec* r)
	{
		return apr_pstrdup(r->pool, r->method);
	}

	static const char* getChildPID(request_rec* r)
	{
		return apr_psprintf(r->pool, "%" APR_PID_T_FMT, getpid());
	}

	static const char* getServerPort(request_rec* r)
	{
		apr_psprintf(r->pool, "%u", r->server->port ? r->server->port : ap_default_port(r));
		return "";
	}

	static const char* getReferer(request_rec* r)
	{
		const char *tempref = apr_table_get(r->headers_in, "Referer");
		return tempref ? tempref : "";
	}

	static const char* getRequestLine(request_rec* r)
	{
		/* NOTE: If the original request contained a password, we
		* re-write the request line here to contain XXXXXX instead:
		* (note the truncation before the protocol string for HTTP/0.9 requests)
		* (note also that r->the_request contains the unmodified request)
		*/
		return (r->parsed_uri.password) ? apr_pstrcat(r->pool, r->method, " ",
														apr_uri_unparse(r->pool, &r->parsed_uri, 0),
														r->assbackwards ? NULL : " ",
														r->protocol, NULL)
														: r->the_request;
	}

	static const char* getStatus(request_rec* r)
	{
		if( r->status <= 0 )
			return "";
		else
			return apr_psprintf(r->pool, "%d", r->status);
	}

	static const char* getRequestTime(request_rec* r)
	{
		return apr_psprintf(r->pool, "%"APR_TIME_T_FMT, apr_time_sec(r->request_time));
	}

	static const char* getRemoteUser(request_rec* r)
	{
		return r->user ? r->user : "";
	}

	static const char* getRequestURI(request_rec* r)
	{
		return r->uri;
	}

	static const char* getVirtualHost(request_rec* r)
	{
		return r->server->server_hostname;
	}

	static const char* getServerName(request_rec* r)
	{
		return ap_get_server_name(r);		//Respect UseCanonicalName
	}

	static const char* getBytesIn(request_rec* r)
	{
		return apr_off_t_toa(r->pool, get_module_config(r)->first);
	}

	static const char* getBytesOut(request_rec* r)
	{
		return apr_off_t_toa(r->pool,  get_module_config(r)->second);
	}

	//Format specifier, buffer size, column name, extractor, SQL
	// !!! These are arranged in table order
	const column_t	columns[] =	{	{'t', "RequestTime", getRequestTime, "TIMESTAMP", "Time of request receipt by httpd"},
											{'B', "BytesSent", getBytesSent, "INT UNSIGNED", "Size of response in bytes, excluding HTTP headers"},
											{'D', "ElapsedSeconds", getElapsedSeconds, "INT UNSIGNED", "Request duration in seconds"},
											{'d', "ElapsedMicoSeconds", getElapsedMicroSeconds, "INT UNSIGNED", "Request duration in microseconds"},
											{'P', "ChildPID", getChildPID, "INT", ""},
											{'p', "ServerPort", getServerPort, "INT UNSIGNED", ""},
											{'s', "Status", getStatus, "INT UNSIGNED", ""},
											{'I', "BytesIn", getBytesIn, "INT UNSIGNED", ""},
											{'O', "BytesOut", getBytesOut, "INT UNSIGNED", ""},
											{'A', "Agent", getAgent, "VARCHAR(255)", ""},
											{'a', "RemoteIP", getRemoteIP, "VARCHAR(255)", ""},
											{'f', "RequestFile", getRequestFile, "VARCHAR(255)", ""},
											{'h', "RemoteHost", getRemoteHost, "VARCHAR(255)", ""},
											{'i', "UniqueID", getUniqueID, "VARCHAR(255)", ""},
											{'l', "RemoteLogname", getRemoteLogname, "VARCHAR(255)", ""},
											{'m', "RequestMethod", getRequestMethod, "VARCHAR(255)", ""},
											{'U', "RequestURI", getRequestURI, "TEXT", ""},
											{'H', "RequestProtocol", getRequestProtocol, "VARCHAR(255)", ""},
											{'q', "RequestArgs", getRequestArgs, "TEXT", ""},
											{'R', "Referer", getReferer, "VARCHAR(255)", ""},
											{'r', "RequestLine", getRequestLine, "TEXT", ""},
											{'u', "RemoteUser", getRemoteUser, "VARCHAR(255)", ""},
											{'v', "VirtualHost", getVirtualHost, "VARCHAR(255)", ""},
											{'V', "ServerName", getServerName, "VARCHAR(255)", ""},
											{0x00, "", NULL, "", ""}	//Terminator
										};

}
