/*	Filename: hooks.cc
	Hook handlers for mod_log_dbd
	
	Created	20061104	Brandon Fosdick <bfoz@bfoz.net>
	Copyright Brandon Fosdick 2006 (BSD License)
*/

#include <utility>

#include "mod_log_dbd.h"
#include "hooks.h"
#include "serverconfig.h"

//Apache includes
//	***	These must be local includes to resolve alloc conflict
#include "http_core.h"			//For ap_logio_add_bytes_out
#include "http_connection.h"	//For ap_hook_pre_connection
#include "http_protocol.h"		//For ap_hook_log_transaction

namespace log_dbd
{
	static const char filter_name[] = "LOG_DBD_INPUT_OUTPUT";

	//Convenience functions
	inline byte_count_t* get_module_config(conn_rec* c)
	{
		return static_cast<byte_count_t*>(ap_get_module_config(c->conn_config, &log_dbd_module));
	}

	//Post-config phase hook
	static int post_config(apr_pool_t* p, apr_pool_t* plog, apr_pool_t* ptemp, server_rec* s)
	{
		void *data = NULL;
		const char *userdata_key = "mod_log_dbd_post_config";
		apr_pool_userdata_get(&data, userdata_key, s->process->pool);
		if (data == NULL)
		{
			/* WARNING: This must *not* be apr_pool_userdata_setn(). */
			apr_pool_userdata_set((const void *)1, userdata_key,
			apr_pool_cleanup_null, s->process->pool);
			return OK; /* This would be the first time through */
		}

		while(s != NULL)
		{
			ServerConfig* cfg = get_server_config(s);	//Get the server config struct
			cfg->post_config();
			s = s->next;
		}
		return OK;
	}
	
	//Handle child init
	//	***	The server_rec list isn't walked by core, so it must be walked here
	static void child_init(apr_pool_t* p, server_rec* s)
	{
		TRACE();

		//Loop through all of the server records and initialize each
		while(s != NULL)
		{
			get_server_config(s)->connect();				//Open the DB connection
			s = s->next;
		}
	}

	static int log_transaction(request_rec* r)
	{
		return get_server_config(r)->log_transaction(r);
	}

	static int pre_connection(conn_rec *c, void *csd)
	{
		byte_count_t* bc = static_cast<byte_count_t*>(apr_pcalloc(c->pool, sizeof(*bc)));
		bc->first = bc->second = 0;	//Clear the byte counters

		ap_set_module_config(c->conn_config, &log_dbd_module, bc);

		ap_add_input_filter(filter_name, NULL, NULL, c);
		ap_add_output_filter(filter_name, NULL, NULL, c);

		return OK;
	}

	static void ap_logio_add_bytes_out(conn_rec *c, apr_off_t bytes)
	{
		get_module_config(c)->second += bytes;
	}

	static apr_status_t in_filter(ap_filter_t *f,
													apr_bucket_brigade *bb,
													ap_input_mode_t mode,
													apr_read_type_e block,
													apr_off_t readbytes)
	{
		apr_off_t length;

		const apr_status_t status = ap_get_brigade(f->next, bb, mode, block, readbytes);

		apr_brigade_length(bb, 0, &length);

		if (length > 0)
			get_module_config(f->c)->first += length;

		return status;
	}

	static apr_status_t out_filter(ap_filter_t *f, apr_bucket_brigade *bb)
	{
		apr_bucket *b = APR_BRIGADE_LAST(bb);

		/* End of data, make sure we flush */
		if (APR_BUCKET_IS_EOS(b))
		{
			APR_BUCKET_INSERT_BEFORE(b, apr_bucket_flush_create(f->c->bucket_alloc));
		}

		return ap_pass_brigade(f->next, bb);
	}

	//Register handlers, hooks, etc
	void register_hooks(apr_pool_t* p)
	{
		TRACE();
		//Apache init hooks
		ap_hook_post_config(post_config, NULL, NULL, APR_HOOK_MIDDLE);
		ap_hook_child_init(child_init, NULL, NULL, APR_HOOK_MIDDLE);
		apr_dbd_init(p);		//Init the DBD library
		ap_hook_log_transaction(log_transaction, NULL, NULL, APR_HOOK_MIDDLE);

		ap_hook_pre_connection(pre_connection, NULL, NULL, APR_HOOK_MIDDLE);
		ap_register_input_filter(filter_name, in_filter, NULL, AP_FTYPE_NETWORK);
		ap_register_output_filter(filter_name, out_filter, NULL, AP_FTYPE_NETWORK);

		APR_REGISTER_OPTIONAL_FN(ap_logio_add_bytes_out);
	}
};
