/*	Filename: serverconfig.h
	Server config object and database interface for mod_log_dbd
	
	Created	20050819	Brandon Fosdick <bfoz@bfoz.net>
		Moved from mod_dav_userdir.h

	Copyright Terran Development Co 2005 (BSD License)

	$Id: serverconfig.h,v 1.12 2007/04/19 05:54:40 bfoz Exp $
*/

#ifndef	SERVERCONFIG_H
#define	SERVERCONFIG_H

//System includes
#include <algorithm>
#include <vector>
#include <sys/stat.h>

#include "apr_dbd.h"
#include "apr_strings.h"
#include "apr_uri.h"

#include "apr_pool_base.h"
#include	"mod_log_dbd.h"

#include "columns.h"

namespace log_dbd
{
	typedef	std::string	Path_t;

	//See if the path has a component that refers to a local socket
	//	Anything that doesn't stat as a socket is assumed to be the db name
	static std::string get_socket(char* path)
	{
		if( (path != NULL) && (strlen(path)>1) )
		{
			struct stat sb;
			char *const end = path + strlen(path);	//Point to one past the end of the string
			char* s = end;

			while( s != path )
			{
				*s = 0;	//Hide the '/'
				if( stat(path, &sb) == 0 )
				{
					if( S_ISSOCK(sb.st_mode) )
					{
						if( s != end )
							*s = '/';	//Restore the '/'
						return std::string(path, s-path);	//Found a socket
					}
				}

				char *const a = strrchr(path, '/');	// Back up one path segment
				if( s != end )
					*s = '/';	//Restore the '/'
				s = a;
			}
		}
			return std::string();
	}

	//Convert a DSN into a parameter string for MySQL
	static std::string mysql_unparse(const apr_uri_t& dsn)
	{
		std::vector<std::string> params;

		if(dsn.hostname != NULL)
			params.push_back(std::string("host=")+dsn.hostname);
		if(dsn.user != NULL )
			params.push_back(std::string("user=")+dsn.user);
		if(dsn.password != NULL )
			params.push_back(std::string("pass=")+dsn.password);
		if(dsn.port_str != NULL )
			params.push_back(std::string("port=")+dsn.port_str);
			
		//See if the path has a component that refers to a local socket
		//	Anything that doesn't stat as a socket is assumed to be the schema name
		if( (dsn.path != NULL) && (strlen(dsn.path)>1) )
		{
			struct stat sb;
			char *const end = dsn.path + strlen(dsn.path);	//Point to one past the end of the string
			char* s = end;

			while( s != dsn.path )
			{
				*s = 0;	//Hide the '/'
				if( stat(dsn.path, &sb) == 0 )
				{
					if( S_ISSOCK(sb.st_mode) )
					{
						params.push_back(std::string("sock=")+dsn.path);	//Found a socket
						break;
					}
				}

				char *const a = strrchr(dsn.path, '/');	// Back up one path segment
				if( s != end )
					*s = '/';	//Restore the '/'
				s = a;
			}
			// In MySQL, dbname and schema refer to the same thing
			params.push_back(std::string("dbname=")+(s+1));	//Set the db name
		}
/*
     {"flags", NULL},
     {"fldsz", NULL},
*/
		std::string p;
		std::vector<std::string>::iterator i=params.begin();
		if( i != params.end() )
			p += (*(i++));
		for(; i != params.end(); ++i)
			p += ("," + (*i));
		return p;
	}

	//Convert a DSN into a parameter string for PostgreSQL
	static std::string pgsql_unparse(const apr_uri_t& dsn)
	{
		std::vector<std::string> params;
		std::string s(get_socket(dsn.path));	//Use a local socket?

		if( !(s.empty()) )
			params.push_back(std::string("host=")+s);
		else
			if(dsn.hostname != NULL)	//Try to use the hostname if no socket was given
				params.push_back(std::string("host=")+dsn.hostname);

		if(dsn.user != NULL )
			params.push_back(std::string("user=")+dsn.user);
		if(dsn.password != NULL )
			params.push_back(std::string("password=")+dsn.password);
		if(dsn.port_str != NULL )
			params.push_back(std::string("port=")+dsn.port_str);
		if( (dsn.path != NULL) && (strlen(dsn.path+s.size()) > 1) )
			params.push_back(std::string("dbname=")+(dsn.path+s.size() + 1));

		std::string p;
		std::vector<std::string>::iterator i=params.begin();
		if( i != params.end() )
			p += (*(i++));
		for(; i != params.end(); ++i)
			p += (" " + (*i));
		return p;
	}

	static std::string dsn_unparse(const apr_uri_t& dsn)
	{
		if( strncmp("mysql", dsn.scheme, 5) == 0 )
			return mysql_unparse(dsn);
		else if( strncmp("pgsql", dsn.scheme, 5) == 0 )
			return pgsql_unparse(dsn);
		else if( strncmp("sqlite3", dsn.scheme, 7) == 0 )
		{
		}
		else if( strncmp("sqlite2", dsn.scheme, 7) == 0 )
		{
		}
		else if( strncmp("firebird", dsn.scheme, 8) == 0 )
		{
		}
		return std::string();
	}

	class ServerConfig : public apr::pool_base<ServerConfig>
	{
		typedef	std::ostringstream	query_t;
		typedef	std::vector<const column_t*>	columns_t;

		server_rec*	server_record;	//The server record that this config belongs to
		apr_dbd_driver_t*	driver;
		apr_dbd_t*	db;

		//Configuration parameters
		std::string schema;
		std::string table_name;
		std::string	format;		//Format string for this server config

		bool	config_handled;	//False until at least one handle_config_* has been called
		bool	config_ready;		//False until after successful post_config()
		bool	connected;			//False until after successful connect()
		bool	isPostgres;			//Database is PostgreSQL

		//These aren't valid until after post_config()
		apr_uri_t	dsn;
		columns_t	logColumns;	//The list of fields to be logged

		//These aren't valid until after connect()
		apr_dbd_prepared_t*	stmtAccess;	//Prepared access log query
		char**	stmtArgs;					//Pointer to array of char pointers for statment args
		query_t		query;					//The access log query

	public:
		//public_only defaults to true for safety/security
		ServerConfig(server_rec* s, apr_pool_t* p) : driver(NULL), db(NULL), config_handled(false), config_ready(false), connected(false), stmtAccess(NULL), server_record(s), apr::pool_base<ServerConfig>(p) {}
		~ServerConfig()
		{
			if( (driver != NULL) && (db != NULL) )
				apr_dbd_close(driver, db);
			if( stmtArgs )
				delete stmtArgs;
		}

		//Use the passed ServerConfig to fill in anything that's missing from this config		
		void merge_from(ServerConfig* s)
		{
			TRACE1(PTR_CHAR(s->get_server_hostname()) << " into " << get_server_hostname());
			CDBG_VAR(schema);
			
			if(s == NULL)
				return;
			if( get_pool() == NULL )
				set_pool(s->get_pool());
			if(driver == NULL)
				driver = s->driver;
			if(schema.empty())
				schema = s->schema;
			if(table_name.empty())
				table_name = s->table_name;
			if(format.empty())
				format = s->format;
		}

		//Configuration setters
		const char* set_dsn(std::string s)
		{
			unsigned index = 0;

			apr_uri_parse(get_pool(), s.c_str(), &dsn);	//Parse the DSN as a URI

			//Get the appropriate database driver
			switch( apr_dbd_get_driver(get_pool(), dsn.scheme, (const apr_dbd_driver_t**)&driver) )
			{
				case APR_ENOTIMPL:
					return apr_psprintf(get_pool(), "No driver for %s", dsn.scheme);
				case APR_EDSOOPEN:
					return apr_psprintf(get_pool(), "Couldn't load apr_dbd_%s.so", dsn.scheme);
				case APR_ESYMNOTFOUND:
					return apr_psprintf(get_pool(), "Failed to load apr_dbd_%s_driver", dsn.scheme);
			}
			
			if( dsn.hostname == NULL )
				return apr_psprintf(get_pool(), "No hostname given");

			if( strncmp("mysql", dsn.scheme, 5) == 0 )
			{
				size_t len = strlen(dsn.path);
				if( (0 == len) || ((1 == len) && ('/' == dsn.path[0])) )
					return apr_psprintf(get_pool(), "A schema name must be provided for MySQL");
			}

			// Store the schema name, less any socket info, 
			//  which amounts to the trailing part of dsn.path
			schema = strrchr(dsn.path, '/') + 1;	//Store the schema name, less the leading /

			config_handled = true;
			return NULL;	//Return success
		}
		const char* set_format(std::string s)
		{
			format = s;
			config_handled = true;
			return NULL;	//Return success
		}

		server_rec* get_server_record() const { return server_record; }
		char* get_server_hostname() const
		{
			if(server_record == NULL)
				return NULL;
			else
				return server_record->server_hostname;
		}

		//Has a minum set of config directives been seen?
		bool checkConfigReady()
		{
			if( (dsn.hostname == NULL) || (NULL == driver) )
				return false;
			return true;
		}

		bool post_config()
		{
			if( !config_handled )	// If this config was never set up, it's
				return false;			//  probably the root config, so ignore it

			if( !checkConfigReady() )		//	Don't bother connecting if this config isn't ready
			{
				CERR("Config not ready");
				return false;
			}

			if( apr_dbd_open(driver, get_pool(), dsn_unparse(dsn).c_str(), &db) != APR_SUCCESS )
			{
				CERR("Can't connect to " << dsn.hostname);
				return false;
			}

			// Postgres is a little different, so make it easier to do the appropriate things
			isPostgres = strncmp("pgsql", dsn.scheme, 5) == 0;

			//Parse the format string into a list of columns
			if( format.empty() )		//	If the format string is empty use "NCSA combined"
				format = combinedFormatSpecifiers;
			for(std::string::iterator i = format.begin(); i != format.end(); ++i)
			{
				const column_t* j = columns;
				while( j->formatSpecifier != 0x00 )
				{
					if( (*i) == j->formatSpecifier )
					{
						logColumns.push_back(j);
						break;
					}
					++j;
				}
			}

			//Bail out if nothing to log (only happens for a bad format string)
			if( logColumns.size() == 0 )
			{
				CERR("No columns to log");
				return false;
			}

			//Sort the column list to put it in table order
			sort(logColumns.begin(), logColumns.end());

			//Make the table name from the hostname, but w/o unsafe characters
			table_name = "log_";
			for(unsigned i = 0; server_record->server_hostname[i] != 0x00; ++i)
			{
				switch(server_record->server_hostname[i])
				{
					case '\n':
					case '\r':
					case '.':
					case ',':
					case '\\':
					case 0x1A:
						break;
					default:
						table_name.push_back(server_record->server_hostname[i]);
				}
			}
			//Append the server's port number to the table name, if it has one
			if( server_record->port != 0 )
			{
				//Silly string library
				query_t temp;
				temp << server_record->port;
				table_name.append(temp.str());
			}

			//Does the database have the requested schema?
			if( !hasSchema() )
			{
				//If the DB is PostgreSQL and the schema doesn't exist, try to create it
				if( isPostgres )
				{
					if( !createSchema(schema) )
						return false;
				}
				CERR("Couldn't find schema " << schema);
				return false;
			}

			//Does the schema have the requested table?
			if( !hasTable() )
			{
				if( !createTable() )	//Try to create the table
					return false;
			}

			//Does the table have the requested columns?
			std::vector<std::string> c;
			if( !getColumns(c) )
			{
				CERR("Couldn't get list of columns");
				return false;
			}

			columns_t::iterator i = logColumns.begin();
			std::string LastFoundColumn;
			for(; i != logColumns.end(); ++i)
			{
				//For each required column, search the list of existing columns
				std::vector<std::string>::iterator j = c.begin();
				while( (j != c.end()) && ((*i)->columnName != (*j)) )
					++j;

				if( j == c.end() )	//If the column wasn't found, add it to the table
				{
					query_t q;
					q << "ALTER TABLE " << schema << "." << table_name << " ADD COLUMN \"" << (*i)->columnName << '"';

					std::string sql((*i)->columnSQL);
					if( isPostgres )
					{
						const std::string::size_type i = sql.find(" UNSIGNED");
						if( i != std::string::npos )
							sql.erase(i);
					}
					q << " " << sql;

					if( !LastFoundColumn.empty() && !isPostgres )
						q << " AFTER " << LastFoundColumn;
					int nrows;
					int e = apr_dbd_query(driver, db, &nrows, q.str().c_str());
					if( e != 0 )
					{
						CERR(getpid() << " Couldn't alter table " << table_name << " because " << apr_dbd_error(driver, db, e));
						return false;
					}
				}
				LastFoundColumn = (*i)->columnName;
			}

			apr_dbd_close(driver, db);
			config_ready = true;
			return true;
		}

		bool connect()
		{
			TRACE1(apr_uri_unparse(get_pool(), &dsn, 0));

			if( !config_ready )	//	Don't bother connecting if this config isn't ready
				return false;

			if( apr_dbd_open(driver, get_pool(), dsn_unparse(dsn).c_str(), &db) != APR_SUCCESS )
			{
				CERR("Can't connect to " << dsn.hostname);
				return false;
			}

			//Create a query string from the list of columns
			columns_t::iterator j = logColumns.begin();
			query << "INSERT INTO " << schema << "." << table_name << " (";
			while(j != logColumns.end())
			{
				if( j != logColumns.begin() )
					query << ',';
				query << (isPostgres ? '"' : '`') << (*j++)->columnName << (isPostgres ? '"' : '`');
			}
			query << ") VALUES(";
			for(unsigned i=0; i < logColumns.size(); ++i)
			{
				if( i != 0 )
					query << ',';
				if( 't'==logColumns[i]->formatSpecifier )
				{
					if( isPostgres )
						query << "ABSTIME(%d)";
					else
						query << "FROM_UNIXTIME(?)";
				}
				else
					query << '%' << ( (logColumns[i]->columnSQL[0] == 'I') ? 'd' : 's');
			}
			query << ")";

			stmtArgs = new(get_pool()) char*[logColumns.size()];	//Allocate argument buffer

			//Prepare the query
			// NOTE: In PostgreSQL the statement name is automatically lower-cased by a bug in DBD
			//			Consequently, passing a name with any upper-case characters will result in an 
			//			error stating that the statement doesn't exist. Use all lower-case.
			int e;
			if( (e = prepare(query, "access", &stmtAccess)) != 0 )
			{
				CERR("Couldn't prepare query: " << query.str() );
				CERR(apr_dbd_error(driver, db, e));
				return false;
			}
			connected = true;	//Ready for action
			return true;
		}

		int log_transaction(request_rec* r)
		{
			if( (NULL == r) || (!connected) || (NULL == stmtAccess) )
				return DECLINED;

			columns_t::iterator i = logColumns.begin();
			for(unsigned j=0; i != logColumns.end(); ++i, ++j)
			{
				if( NULL != (*i)->get )
					stmtArgs[j] = (char *)(*i)->get(r);
				if( NULL== stmtArgs[j])
					CERR("Bad argument for " << ((*i)->columnName));
			}

			byte_count_t* bc = get_module_config(r);
			if( bc != NULL )
				bc->first = bc->second = 0;	//Clear the byte counters

			int e;
			if( e = pquery(r->pool, stmtAccess, logColumns.size(), (const char**)stmtArgs) != 0 )
			{
				CERR("Couldn't submit query" << ": " << apr_dbd_error(driver, db, e));
				return DECLINED;
			}

			return OK;
		}

	private:

		//Check for schema existence
		bool hasSchema()
		{
			query_t query;
			query << "SELECT 1 FROM information_schema.schemata WHERE SCHEMA_NAME = '" << schema << "'";

 			apr_dbd_results_t* r = NULL;
			// !!! The last argument must be 1 for apr_dbd_num_tuples() to work
			int e = apr_dbd_select(driver, get_pool(), db, &r, query.str().c_str(), 1);
			if( e != 0 )
			{
				CERR("Couldn't find schema " << schema << " because " << apr_dbd_error(driver, db, e));
				return false;
			}

			if( r == NULL )
			{
				CERR("No results for schema " << schema << " because " << apr_dbd_error(driver, db, e));
				return false;
			}

			return (apr_dbd_num_tuples(driver, r) > 0);
		}

		// Try to create a schema in the current database
		//	!!! PostgreSQL only	!!!
		bool createSchema(std::string& s)
		{
			query_t query;
			query << "CREATE SCHEMA " << schema;

			int nrows;
			int e = apr_dbd_query(driver, db, &nrows, query.str().c_str());
			if( e != 0 )
			{
				CERR("Couldn't create schema " << s << " because " << apr_dbd_error(driver, db, e));
				return false;
			}
			return true;
		}

		//Check for schema existence
		bool hasTable()
		{
			query_t query;
			query << "SELECT 1 FROM information_schema.tables WHERE TABLE_SCHEMA='" << schema << "' AND TABLE_NAME='" << table_name << "'";

 			apr_dbd_results_t* r = NULL;
			// !!! The last argument must be 1 for apr_dbd_num_tuples() to work
			int e = apr_dbd_select(driver, get_pool(), db, &r, query.str().c_str(), 1);
			if( e != 0 )
			{
				CERR("Couldn't find table " << table_name << " because " << apr_dbd_error(driver, db, e));
				return false;
			}

			if( r == NULL )
			{
				CERR("No results for table " << table_name << " because " << apr_dbd_error(driver, db, e));
				return false;
			}

			return (apr_dbd_num_tuples(driver, r) > 0);
		}

		//Create a table
		bool createTable()
		{
			query_t query;
			query << "CREATE TABLE " << schema << "." << table_name << " (";

			columns_t::iterator j = logColumns.begin();
			for(; j != logColumns.end(); ++j)
			{
				if( j != logColumns.begin() )
					query << ',';
				query << (isPostgres ? '"' : '`') << (*j)->columnName << (isPostgres ? '"' : '`');
				std::string sql((*j)->columnSQL);
				if( isPostgres )
				{
					const std::string::size_type i = sql.find(" UNSIGNED");
					if( i != std::string::npos )
						sql.erase(i);
				}
				query << " " << sql;
				if( !isPostgres )
					query << " COMMENT '" << (*j)->columnComment << "'";
			}
			query << ")";

			int nrows;
			int e = apr_dbd_query(driver, db, &nrows, query.str().c_str());
			if( e != 0 )
			{
				CERR("Couldn't create table " << table_name << " because " << apr_dbd_error(driver, db, e));
				return false;
			}
			return true;
		}

		bool getColumns(std::vector<std::string>& c)
		{
			int a;
			query_t q;
			q << "SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA='" << schema << "' AND TABLE_NAME='" << table_name << "' ORDER BY ORDINAL_POSITION ASC";

			apr_dbd_results_t* r = NULL;
			int e = apr_dbd_select(driver, get_pool(), db, &r, q.str().c_str(), 0);
			if( e != 0 )
			{
				CERR("Couldn't get columns because " << apr_dbd_error(driver, db, e));
				return false;
			}

			apr_dbd_row_t *row = NULL;
			while( apr_dbd_get_row(driver, get_pool(), r, &row, -1) != -1 )
			{
				c.push_back(apr_dbd_get_entry(driver, row, 0));
			}
			return true;
		}

		//Shortcuts for doing queries
		//	*** Caller is responsible for freeing results

		const char* error(int errnum)
		{
			return apr_dbd_error(driver, db, errnum);
		}

		//Prepare a statement
		int prepare(query_t &query, const char *label, apr_dbd_prepared_t **statement)
		{
			return prepare(query.str().c_str(), label, statement);
		}
		int prepare(const char* query, const char *label, apr_dbd_prepared_t **statement)
		{
			return apr_dbd_prepare(driver, get_pool(), db, query, label, statement);
		}

		//Run a prepared query that doesn't return a result set
		// The passed pool must be a request pool and not the server pool. For some
		//  reason using the server pool will cause a segfault in apr_dbd_pquery().
		int pquery(apr_pool_t *p, apr_dbd_prepared_t *statement, int nargs, const char **args)
		{
			int nrows;
			return apr_dbd_pquery(driver, p, db, &nrows, statement, nargs, args);
		}
	};

	//Convenience functions
	inline ServerConfig* get_server_config(server_rec* s)
	{
		return static_cast<ServerConfig*>(ap_get_module_config(s->module_config, &log_dbd_module));
	}

	inline ServerConfig* get_server_config(request_rec* r)
	{
		return get_server_config(r->server);
	}

	inline ServerConfig* get_server_config(cmd_parms* r)
	{
		return get_server_config(r->server);
	}

	//This really sucks. Why does apache have to use so many damn void pointers?
	inline ServerConfig* get_server_config(void* p)
	{
		return static_cast<ServerConfig*>(p);
	}
}

#endif
